from pyaprilaire.const import Action, Attribute, FunctionalDomain
from pyaprilaire.packet import NackPacket, Packet


def test_invalid_action():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 7, 1, 1, 0]))

    assert len(packets) == 0


def test_invalid_functional_domain():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 3, 17, 1, 0]))

    assert len(packets) == 0


def test_unmapped():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 3, 13, 1, 0]))

    assert len(packets) == 0


def test_packet_empty_parse():
    packets = list(Packet.parse(bytes()))

    assert len(packets) == 0


def test_packet_single_parse():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

    assert len(packets) == 1


def test_packet_single_action():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

    packet = packets[0]

    assert packet.action == Action.READ_REQUEST


def test_packet_single_functional_domain():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

    packet = packets[0]

    assert packet.functional_domain == FunctionalDomain.SETUP


def test_packet_single_attribute():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

    packet = packets[0]

    assert packet.attribute == 1


def test_packet_single_extra_data():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 8, 3, 2, 1, 1, 1, 1, 1, 10, 146])
    )

    assert len(packets) == 1


def test_packet_multiple_parse():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
    )

    assert len(packets) == 2


def test_packet_multiple_action():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
    )

    packet = packets[1]

    assert packet.action == Action.READ_RESPONSE


def test_packet_multiple_functional_domain():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
    )

    packet = packets[1]

    assert packet.functional_domain == FunctionalDomain.SCHEDULING


def test_packet_multiple_attribute():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
    )

    packet = packets[1]

    assert packet.attribute == 4


def test_nack_parse():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 2, 6, 1, 0]))

    assert len(packets) == 1


def test_nack_and_packet_parse():
    packets: list[Packet] = list(
        Packet.parse(
            [
                0x01,
                0x04,
                0x00,
                0x02,
                0x06,
                0x03,
                0xCD,
                0x01,
                0x01,
                0x00,
                0x11,
                0x03,
                0x08,
                0x02,
                0xB4,
                0x82,
                0x55,
                0x50,
                0x93,
                0x6D,
                0x01,
                0x49,
                0x02,
                0x01,
                0x02,
                0x0D,
                0x04,
                0x0E,
                0x51,
            ]
        )
    )

    assert len(packets) == 2


def test_nack_packet_parse():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 6, 1, 0]))

    assert isinstance(packets[0], NackPacket)


def test_control_1_parse():
    packets: list[Packet] = list(Packet.parse([1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107]))

    packet = packets[0]

    assert packet.data == {
        Attribute.MODE: 1,
        Attribute.FAN_MODE: 2,
        Attribute.HEAT_SETPOINT: 10,
        Attribute.COOL_SETPOINT: 20,
    }


def test_scheduling_4_parse():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 13, 3, 3, 4, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 42])
    )

    packet = packets[0]

    assert packet.data == {
        Attribute.HOLD: 1,
    }


def test_sensor_2_parse():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 11, 3, 5, 2, 1, 10, 2, 20, 3, 50, 4, 60, 12])
    )

    packet = packets[0]

    assert packet.data == {
        Attribute.INDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS: 1,
        Attribute.INDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE: 10,
        Attribute.OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS: 2,
        Attribute.OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE: 20,
        Attribute.INDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS: 3,
        Attribute.INDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE: 50,
        Attribute.OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS: 4,
        Attribute.OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE: 60,
    }


def test_identification_2_parse():
    packets: list[Packet] = list(
        Packet.parse([1, 1, 0, 9, 3, 8, 2, 1, 2, 3, 4, 5, 6, 176])
    )

    packet = packets[0]

    assert packet.data == {Attribute.MAC_ADDRESS: "1:2:3:4:5:6"}


def test_identification_4_parse():
    packets: list[Packet] = list(
        Packet.parse(
            [
                1,
                1,
                0,
                27,
                3,
                8,
                4,
                49,
                50,
                51,
                52,
                53,
                0,
                0,
                0,
                84,
                101,
                115,
                116,
                32,
                78,
                97,
                109,
                101,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                180,
            ]
        )
    )

    packet = packets[0]

    assert packet.data == {Attribute.LOCATION: "12345", Attribute.NAME: "Test Name"}


def test_decode_temperature():
    temperature = Packet._decode_temperature(0x15)

    assert temperature == 21


def test_decode_temperature_negative():
    temperature = Packet._decode_temperature(0x95)

    assert temperature == -21


def test_decode_temperature_decimal():
    temperature = Packet._decode_temperature(0x5A)

    assert temperature == 26.5


def test_decode_temperature_negative_decimal():
    temperature = Packet._decode_temperature(0xDA)

    assert temperature == -26.5


def test_decode_humidity_zero():
    humidity = Packet._decode_humidity(0)

    assert humidity == 0


def test_decode_humidity_1():
    humidity = Packet._decode_humidity(1)

    assert humidity == 1


def test_decode_humidity_99():
    humidity = Packet._decode_humidity(99)

    assert humidity == 99


def test_decode_humidity_100():
    humidity = Packet._decode_humidity(100)

    assert humidity == 100


def test_decode_humidity_nonzero():
    humidity = Packet._decode_humidity(50)

    assert humidity == 50


def test_decode_humidity_negative():
    humidity = Packet._decode_humidity(-50)

    assert humidity == None


def test_encode_temperature():
    encoded_temperature = Packet._encode_temperature(21)

    assert encoded_temperature == 0x15


def test_encode_temperature_fraction():
    encoded_temperature = Packet._encode_temperature(26.5)

    assert encoded_temperature == 0x5A


def test_encode_temperature_negative():
    encoded_temperature = Packet._encode_temperature(-21)

    assert encoded_temperature == 0x95


def test_encode_temperature_negative_fraction():
    encoded_temperature = Packet._encode_temperature(-26.5)

    assert encoded_temperature == 0xDA


def test_serialize_nack():
    serialized = NackPacket(2).serialize()

    assert serialized == bytes([1, 0, 0, 2, 6, 2, 227])


def test_serialize_raw():
    serialized = Packet(
        Action.READ_REQUEST, FunctionalDomain.CONTROL, 1, 1, 1, raw_data=[1, 2, 3]
    ).serialize()

    assert serialized == bytes([1, 1, 0, 6, 2, 2, 1, 1, 2, 3, 133])


def test_serialize_single_packet_no_data():
    serialized = Packet(
        Action.READ_REQUEST,
        FunctionalDomain.CONTROL,
        1,
        1,
        1,
    ).serialize()

    assert serialized == bytes([1, 1, 0, 3, 2, 2, 1, 70])


def test_serialize_single_packet_no_data():
    serialized = Packet(
        Action.READ_REQUEST,
        FunctionalDomain.CONTROL,
        1,
        1,
        1,
    ).serialize()

    assert serialized == bytes([1, 1, 0, 3, 2, 2, 1, 70])


def test_control_1_serialize():
    serialized = Packet(
        Action.READ_RESPONSE,
        FunctionalDomain.CONTROL,
        1,
        1,
        1,
        data={
            Attribute.MODE: 1,
            Attribute.FAN_MODE: 2,
            Attribute.HEAT_SETPOINT: 10,
            Attribute.COOL_SETPOINT: 20,
        },
    ).serialize()

    assert serialized == bytes([1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107])


def test_scheduling_4_serialize():
    serialized = Packet(
        Action.READ_RESPONSE,
        FunctionalDomain.SCHEDULING,
        4,
        1,
        1,
        data={
            Attribute.HOLD: 1,
        },
    ).serialize()

    assert serialized == bytes([1, 1, 0, 13, 3, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55])


def test_sensor_2_serialize():
    serialized = Packet(
        Action.READ_RESPONSE,
        FunctionalDomain.SENSORS,
        2,
        1,
        1,
        data={
            Attribute.INDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS: 1,
            Attribute.INDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE: 10,
            Attribute.OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS: 2,
            Attribute.OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE: 20,
            Attribute.INDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS: 3,
            Attribute.INDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE: 50,
            Attribute.OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS: 4,
            Attribute.OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE: 60,
        },
    ).serialize()

    assert serialized == bytes([1, 1, 0, 11, 3, 5, 2, 1, 10, 2, 20, 3, 50, 4, 60, 12])


def test_identification_2_serialize():
    serialized = Packet(
        Action.READ_RESPONSE,
        FunctionalDomain.IDENTIFICATION,
        2,
        1,
        1,
        data={Attribute.MAC_ADDRESS: [1, 2, 3, 4, 5, 6]},
    ).serialize()

    assert serialized == bytes([1, 1, 0, 9, 3, 8, 2, 1, 2, 3, 4, 5, 6, 176])


def test_identification_4_serialize():
    serialized = Packet(
        Action.READ_RESPONSE,
        FunctionalDomain.IDENTIFICATION,
        4,
        1,
        1,
        data={Attribute.LOCATION: "12345", Attribute.NAME: "Test Name"},
    ).serialize()

    assert serialized == bytes(
        [
            1,
            1,
            0,
            27,
            3,
            8,
            4,
            49,
            50,
            51,
            52,
            53,
            0,
            0,
            0,
            84,
            101,
            115,
            116,
            32,
            78,
            97,
            109,
            101,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            180,
        ]
    )



"""Constants for the Aprilaire integration"""

from __future__ import annotations

from enum import Enum, IntEnum

try:
    from enum import StrEnum
except:

    class StrEnum(str, Enum):
        pass


class Action(IntEnum):
    """An action for commands"""

    NONE = 0
    WRITE = 1
    READ_REQUEST = 2
    READ_RESPONSE = 3
    COS = 5
    NACK = 6


class FunctionalDomain(IntEnum):
    """A functional domain for commands"""

    NONE = 0
    SETUP = 1
    CONTROL = 2
    SCHEDULING = 3
    ALERTS = 4
    SENSORS = 5
    LOCKOUT = 6
    STATUS = 7
    IDENTIFICATION = 8
    MESSAGING = 9
    DISPLAY = 10
    WEATHER = 13
    FIRMWARE_UPDATE = 14
    DEBUG_COMMANDS = 15
    NACK = 16


# Known model numbers per Aprilaire. Additional models may be discovered.
MODELS = {
    0: "8476W",
    1: "8810",
    2: "8620W",
    3: "8820",
    4: "8910W",
    5: "8830",
    6: "8920W",
    7: "8840",
    28: "6045M",
}

QUEUE_FREQUENCY = 0.5


class Attribute(StrEnum):
    ERROR = "error"
    AVAILABLE = "available"
    CONNECTED = "connected"
    CONNECTING = "connecting"
    RECONNECTING = "reconnecting"
    STOPPED = "stopped"

    AWAY_AVAILABLE = "away_available"
    MODE = "mode"
    FAN_MODE = "fan_mode"
    HEAT_SETPOINT = "heat_setpoint"
    COOL_SETPOINT = "cool_setpoint"
    DEHUMIDIFICATION_SETPOINT = "dehumidification_setpoint"
    HUMIDIFICATION_SETPOINT = "humidification_setpoint"
    FRESH_AIR_MODE = "fresh_air_mode"
    FRESH_AIR_EVENT = "fresh_air_event"
    AIR_CLEANING_MODE = "air_cleaning_mode"
    AIR_CLEANING_EVENT = "air_cleaning_event"
    THERMOSTAT_MODES = "thermostat_modes"
    AIR_CLEANING_AVAILABLE = "air_cleaning_available"
    VENTILATION_AVAILABLE = "ventilation_available"
    DEHUMIDIFICATION_AVAILABLE = "dehumidification_available"
    HUMIDIFICATION_AVAILABLE = "humidification_available"
    HOLD = "hold"
    BUILT_IN_TEMPERATURE_SENSOR_STATUS = "built_in_temperature_sensor_status"
    BUILT_IN_TEMPERATURE_SENSOR_VALUE = "built_in_temperature_sensor_value"
    WIRED_REMOTE_TEMPERATURE_SENSOR_STATUS = "wired_remote_temperature_sensor_status"
    WIRED_REMOTE_TEMPERATURE_SENSOR_VALUE = "wired_remote_temperature_sensor_value"
    WIRED_OUTDOOR_TEMPERATURE_SENSOR_STATUS = "wired_outdoor_temperature_sensor_status"
    WIRED_OUTDOOR_TEMPERATURE_SENSOR_VALUE = "wired_outdoor_temperature_sensor_value"
    BUILT_IN_HUMIDITY_SENSOR_STATUS = "built_in_humidity_sensor_status"
    BUILT_IN_HUMIDITY_SENSOR_VALUE = "built_in_humidity_sensor_value"
    RAT_SENSOR_STATUS = "rat_sensor_status"
    RAT_SENSOR_VALUE = "rat_sensor_value"
    LAT_SENSOR_STATUS = "lat_sensor_status"
    LAT_SENSOR_VALUE = "lat_sensor_value"
    WIRELESS_OUTDOOR_TEMPERATURE_SENSOR_STATUS = (
        "wireless_outdoor_temperature_sensor_status"
    )
    WIRELESS_OUTDOOR_TEMPERATURE_SENSOR_VALUE = (
        "wireless_outdoor_temperature_sensor_value"
    )
    WIRELESS_OUTDOOR_HUMIDITY_SENSOR_STATUS = "wireless_outdoor_humidity_sensor_status"
    WIRELESS_OUTDOOR_HUMIDITY_SENSOR_VALUE = "wireless_outdoor_humidity_sensor_value"
    INDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS = (
        "indoor_temperature_controlling_sensor_status"
    )
    INDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE = (
        "indoor_temperature_controlling_sensor_value"
    )
    OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_STATUS = (
        "outdoor_temperature_controlling_sensor_status"
    )
    OUTDOOR_TEMPERATURE_CONTROLLING_SENSOR_VALUE = (
        "outdoor_temperature_controlling_sensor_value"
    )
    INDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS = (
        "indoor_humidity_controlling_sensor_status"
    )
    INDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE = (
        "indoor_humidity_controlling_sensor_value"
    )
    OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_STATUS = (
        "outdoor_humidity_controlling_sensor_status"
    )
    OUTDOOR_HUMIDITY_CONTROLLING_SENSOR_VALUE = (
        "outdoor_humidity_controlling_sensor_value"
    )
    SYNCED = "synced"
    HEATING_EQUIPMENT_STATUS = "heating_equipment_status"
    COOLING_EQUIPMENT_STATUS = "cooling_equipment_status"
    PROGRESSIVE_RECOVERY = "progressive_recovery"
    FAN_STATUS = "fan_status"
    DEHUMIDIFICATION_STATUS = "dehumidification_status"
    HUMIDIFICATION_STATUS = "humidification_status"
    VENTILATION_STATUS = "ventilation_status"
    AIR_CLEANING_STATUS = "air_cleaning_status"
    HARDWARE_REVISION = "hardware_revision"
    FIRMWARE_MAJOR_REVISION = "firmware_major_revision"
    FIRMWARE_MINOR_REVISION = "firmware_minor_revision"
    PROTOCOL_MAJOR_REVISION = "protocol_major_revision"
    MODEL_NUMBER = "model_number"
    GAINSPAN_FIRMWARE_MAJOR_REVISION = "gainspan_firmware_major_revision"
    GAINSPAN_FIRMWARE_MINOR_REVISION = "gainspan_firmware_minor_revision"
    MAC_ADDRESS = "mac_address"
    LOCATION = "location"
    NAME = "name"
    OUTDOOR_SENSOR_STATUS = "outdoor_sensor_status"
    OUTDOOR_SENSOR = "outdoor_sensor"


import asyncio
import logging
import tracemalloc
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyaprilaire.client import AprilaireClient, _AprilaireClientProtocol
from pyaprilaire.const import Action, Attribute, FunctionalDomain
from pyaprilaire.packet import Packet

tracemalloc.start()


@pytest.fixture
def logger():
    logger = logging.getLogger()
    logger.propagate = False

    return logger


@pytest.fixture
def protocol(event_loop, logger):
    data_received_callback = AsyncMock()
    reconnect_action = AsyncMock()

    return _AprilaireClientProtocol(data_received_callback, reconnect_action, logger)


@pytest.fixture
def client(event_loop, logger, protocol):
    data_received_callback = Mock()

    client = AprilaireClient(None, None, data_received_callback, logger, 10, 10)

    client.protocol = protocol

    return client


def test_protocol_connection_made(protocol: _AprilaireClientProtocol):
    protocol._queue_loop = AsyncMock()
    protocol._update_status = AsyncMock()

    protocol.connection_made(None)

    assert protocol._queue_loop.call_count == 1
    assert protocol._update_status.call_count == 1


async def test_protocol_update_status(protocol: _AprilaireClientProtocol):
    sleep_mock = AsyncMock()

    with patch("asyncio.sleep", new=sleep_mock):
        await protocol._update_status()

    assert protocol.packet_queue.qsize() == 9
    assert sleep_mock.call_count == 1


async def test_protocol_queue_loop(protocol: _AprilaireClientProtocol):
    await protocol.read_control()
    await protocol.read_scheduling()

    sleep_mock = AsyncMock()
    protocol.transport = Mock(asyncio.Transport)

    with patch("asyncio.sleep", new=sleep_mock):
        await protocol._queue_loop(loop_count=1)

    assert sleep_mock.call_count == 1
    assert protocol.transport.write.call_count == 2


def test_protocol_data_received(protocol: _AprilaireClientProtocol):
    protocol.data_received(bytes([1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107]))

    assert protocol.data_received_callback.call_count == 1

    (functional_domain, attribute, data) = protocol.data_received_callback.call_args[0]

    assert functional_domain == FunctionalDomain.CONTROL
    assert attribute == 1
    assert data == {
        Attribute.MODE: 1,
        Attribute.FAN_MODE: 2,
        Attribute.HEAT_SETPOINT: 10,
        Attribute.COOL_SETPOINT: 20,
    }


def test_protocol_data_received_nack(protocol: _AprilaireClientProtocol):
    protocol.data_received(bytes([1, 1, 0, 2, 6, 1, 0]))

    assert protocol.data_received_callback.call_count == 0


def test_protocol_data_received_error(protocol: _AprilaireClientProtocol):
    protocol.data_received(bytes([1, 1, 0, 4, 3, 7, 8, 2, 149]))

    assert protocol.data_received_callback.call_count == 1

    (functional_domain, attribute, data) = protocol.data_received_callback.call_args[0]

    assert functional_domain == FunctionalDomain.STATUS
    assert attribute == 8
    assert data == {
        "error": 2,
    }


def test_protocol_mode_re_read(protocol: _AprilaireClientProtocol):
    protocol.read_control = AsyncMock()

    protocol.data_received(bytes([1, 1, 0, 7, 5, 2, 1, 1, 2, 10, 20, 127]))

    assert protocol.read_control.call_count == 1


def test_protocol_connection_lost(protocol: _AprilaireClientProtocol):
    protocol.connection_lost(None)

    assert protocol.data_received_callback.call_count == 1

    (functional_domain, attribute, data) = protocol.data_received_callback.call_args[0]

    assert functional_domain == FunctionalDomain.NONE
    assert attribute == 0
    assert data == {
        "available": False,
    }


def test_protocol_get_sequence(protocol: _AprilaireClientProtocol):
    sequence = protocol._get_sequence()
    assert sequence == 1

    sequence = protocol._get_sequence()
    assert sequence == 2

    protocol.sequence = 127
    sequence = protocol._get_sequence()
    assert sequence == 0

    sequence = protocol._get_sequence()
    assert sequence == 1


async def test_protocol_send_packet(protocol: _AprilaireClientProtocol):
    protocol.packet_queue.put = AsyncMock()
    protocol.packet_queue.put_nowait = Mock()

    original_packet = Packet(
        Action.WRITE,
        FunctionalDomain.CONTROL,
        1,
        data={
            Attribute.MODE: 1,
            Attribute.FAN_MODE: 0,
            Attribute.HEAT_SETPOINT: 0,
            Attribute.COOL_SETPOINT: 0,
        },
    )

    await protocol._send_packet(original_packet)

    assert protocol.packet_queue.put.call_count == 1

    (sent_packet) = protocol.packet_queue.put.call_args[0][0]

    assert original_packet == sent_packet


def assertPacketQueueContains(protocol: _AprilaireClientProtocol, packet: Packet):
    queue_items = list(protocol.packet_queue._queue)

    assert any(qp == packet for qp in queue_items) == True


async def test_protocol_read_sensors(protocol: _AprilaireClientProtocol):
    await protocol.read_sensors()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
    )


async def test_protocol_read_control(protocol: _AprilaireClientProtocol):
    await protocol.read_control()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
    )


async def test_protocol_read_scheduling(protocol: _AprilaireClientProtocol):
    await protocol.read_scheduling()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
    )


async def test_protocol_update_mode(protocol: _AprilaireClientProtocol):
    await protocol.update_mode(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 1,
                Attribute.FAN_MODE: 0,
                Attribute.HEAT_SETPOINT: 0,
                Attribute.COOL_SETPOINT: 0,
            },
        ),
    )


async def test_protocol_update_fan_mode(protocol: _AprilaireClientProtocol):
    await protocol.update_fan_mode(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 0,
                Attribute.FAN_MODE: 1,
                Attribute.HEAT_SETPOINT: 0,
                Attribute.COOL_SETPOINT: 0,
            },
        ),
    )


async def test_protocol_update_setpoint(protocol: _AprilaireClientProtocol):
    await protocol.update_setpoint(10, 20)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 0,
                Attribute.FAN_MODE: 0,
                Attribute.HEAT_SETPOINT: 20,
                Attribute.COOL_SETPOINT: 10,
            },
        ),
    )


async def test_protocol_set_hold(protocol: _AprilaireClientProtocol):
    await protocol.set_hold(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.SCHEDULING,
            4,
            data={
                Attribute.HOLD: 1,
            },
        ),
    )


async def test_protocol_sync(protocol: _AprilaireClientProtocol):
    await protocol.sync()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.STATUS,
            2,
            data={
                Attribute.SYNCED: 1,
            },
        ),
    )


async def test_protocol_configure_cos(protocol: _AprilaireClientProtocol):
    pass


async def test_protocol_read_mac_address(protocol: _AprilaireClientProtocol):
    await protocol.read_mac_address()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            2,
        ),
    )


async def test_protocol_read_thermostat_status(protocol: _AprilaireClientProtocol):
    await protocol.read_thermostat_status()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.CONTROL,
            7,
        ),
    )


async def test_protocol_read_thermostat_name(protocol: _AprilaireClientProtocol):
    await protocol.read_thermostat_name()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            5,
        ),
    )


def test_protocol_empty_packet_queue(protocol: _AprilaireClientProtocol):
    protocol.packet_queue.put_nowait({})
    protocol.packet_queue.put_nowait({})

    assert protocol.packet_queue.qsize() == 2

    protocol._empty_packet_queue()

    assert protocol.packet_queue.qsize() == 0


def test_protocol_empty_packet_queue_error(protocol: _AprilaireClientProtocol):
    protocol.packet_queue.get_nowait = Mock(side_effect=Exception)

    protocol.packet_queue.put_nowait({})
    protocol.packet_queue.put_nowait({})

    assert protocol.packet_queue.qsize() == 2

    protocol._empty_packet_queue()

    assert protocol.packet_queue.qsize() == 2


def test_client_create_protocol(client: AprilaireClient):
    protocol = client.create_protocol()

    assert isinstance(protocol, _AprilaireClientProtocol)


async def test_client_data_received(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    functional_domain = FunctionalDomain.CONTROL
    attribute = 1
    data = {"testKey": "testValue"}

    loop = asyncio.get_event_loop()
    future = loop.create_future()

    future_key = (functional_domain, attribute)

    if future_key not in client.futures:
        client.futures[future_key] = []

    client.futures[future_key].append(future)

    await client.data_received(functional_domain, attribute, data)

    assert client.data_received_callback.call_count == 1
    assert client.data_received_callback.call_args[0][0] == data

    assert future.result() == data


async def test_client_data_received_empty(client: AprilaireClient):
    await client.data_received(None, None, None)


async def test_client_data_received_state_error(client: AprilaireClient):
    functional_domain = FunctionalDomain.CONTROL
    attribute = 1

    loop = asyncio.get_event_loop()
    future = loop.create_future()

    future_key = (functional_domain, attribute)

    if future_key not in client.futures:
        client.futures[future_key] = []

    client.futures[future_key].append(future)

    future.set_result({})

    await client.data_received(functional_domain, attribute, {})


def test_client_state_changed(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    client.connected = True
    client.stopped = True
    client.reconnecting = True

    client.state_changed()

    assert client.data_received_callback.call_count == 1
    assert client.data_received_callback.call_args[0][0] == {
        Attribute.CONNECTED: True,
        Attribute.STOPPED: True,
        Attribute.RECONNECTING: True,
    }


def assertPacketQueueContains(protocol: _AprilaireClientProtocol, packet: Packet):
    queue_items = list(protocol.packet_queue._queue)

    assert any(qp == packet for qp in queue_items) == True


async def test_client_read_sensors(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_sensors()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
    )


async def test_client_read_control(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_control()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
    )


async def test_client_read_scheduling(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_scheduling()

    assertPacketQueueContains(
        protocol, Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
    )


async def test_client_update_mode(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.update_mode(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 1,
                Attribute.FAN_MODE: 0,
                Attribute.HEAT_SETPOINT: 0,
                Attribute.COOL_SETPOINT: 0,
            },
        ),
    )


async def test_client_update_fan_mode(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.update_fan_mode(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 0,
                Attribute.FAN_MODE: 1,
                Attribute.HEAT_SETPOINT: 0,
                Attribute.COOL_SETPOINT: 0,
            },
        ),
    )


async def test_client_update_setpoint(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.update_setpoint(10, 20)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                Attribute.MODE: 0,
                Attribute.FAN_MODE: 0,
                Attribute.HEAT_SETPOINT: 20,
                Attribute.COOL_SETPOINT: 10,
            },
        ),
    )


async def test_client_set_hold(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_hold(1)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.SCHEDULING,
            4,
            data={
                Attribute.HOLD: 1,
            },
        ),
    )


async def test_client_set_dehumidification_setpoint(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_dehumidification_setpoint(50)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            3,
            data={
                Attribute.DEHUMIDIFICATION_SETPOINT: 50,
            },
        ),
    )


async def test_client_set_humidification_setpoint(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_humidification_setpoint(50)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            4,
            data={
                Attribute.HUMIDIFICATION_SETPOINT: 50,
            },
        ),
    )


async def test_client_set_fresh_air(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_fresh_air(1, 3)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            5,
            data={
                Attribute.FRESH_AIR_MODE: 1,
                Attribute.FRESH_AIR_EVENT: 3,
            },
        ),
    )


async def test_client_set_air_cleaning(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_air_cleaning(1, 3)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            6,
            data={
                Attribute.AIR_CLEANING_MODE: 1,
                Attribute.AIR_CLEANING_EVENT: 3,
            },
        ),
    )


async def test_client_sync(client: AprilaireClient, protocol: _AprilaireClientProtocol):
    await client.sync()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.STATUS,
            2,
            data={
                Attribute.SYNCED: 1,
            },
        ),
    )


async def test_client_configure_cos(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    pass


async def test_client_read_mac_address(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_mac_address()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            2,
        ),
    )


async def test_client_read_thermostat_status(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_thermostat_status()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.CONTROL,
            7,
        ),
    )


async def test_client_read_thermostat_name(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.read_thermostat_name()

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            5,
        ),
    )

async def test_client_set_written_outdoor_temperature_value(
    client: AprilaireClient, protocol: _AprilaireClientProtocol
):
    await client.set_written_outdoor_temperature_value(10)

    assertPacketQueueContains(
        protocol,
        Packet(
            Action.WRITE,
            FunctionalDomain.SENSORS,
            4,
            data={
                Attribute.OUTDOOR_SENSOR_STATUS: 0,
                Attribute.OUTDOOR_SENSOR: 10
            }
        )
    )

async def test_client_wait_for_response_success(client: AprilaireClient):
    wait_for_mock = AsyncMock(return_value=True)

    with patch("asyncio.wait_for", new=wait_for_mock):
        wait_for_response_result = await client.wait_for_response(
            FunctionalDomain.CONTROL, 1, 1
        )

    assert wait_for_response_result == True


async def test_client_wait_for_response_timeout(client: AprilaireClient):
    wait_for_mock = AsyncMock(side_effect=asyncio.exceptions.TimeoutError)

    with patch("asyncio.wait_for", new=wait_for_mock):
        wait_for_response_result = await client.wait_for_response(
            FunctionalDomain.CONTROL, 1, 1
        )

    assert wait_for_response_result == None


async def test_client_reconnect_with_delay(client: AprilaireClient):
    reconnect_mock = AsyncMock()

    with patch("pyaprilaire.socket_client.SocketClient._reconnect", new=reconnect_mock):
        await client._reconnect_with_delay()

        assert reconnect_mock.call_count == 1
        assert reconnect_mock.call_args[0][0] == client.retry_connection_interval



# pylint: disable=redefined-outer-name,missing-function-docstring,missing-module-docstring,protected-access
import asyncio
import logging
from asyncio import Protocol
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyaprilaire.client import SocketClient


@pytest.fixture
def logger():
    logger = logging.getLogger()
    logger.propagate = False

    return logger


@pytest.fixture
def client(logger):
    return SocketClient(None, None, None, logger)


def patch_socket(func):
    async def wrapper(client, *args, **kwargs):
        state_changed_mock = Mock()
        create_connection_mock = AsyncMock()
        create_protocol_mock = Mock(spec=Protocol)

        with (
            patch(
                "asyncio.BaseEventLoop.create_connection",
                new=create_connection_mock,
            ),
            patch(
                "pyaprilaire.socket_client.SocketClient.state_changed",
                new=state_changed_mock,
            ),
            patch(
                "pyaprilaire.socket_client.SocketClient.create_protocol",
                new=create_protocol_mock,
            ),
        ):
            await func(client, *args, **kwargs)

    return wrapper


@patch_socket
async def test_client_status(client: SocketClient):
    await client.start_listen()

    assert not client.stopped
    assert client.connected
    assert not client.reconnecting
    assert not client.auto_reconnecting

    client._disconnect()

    assert not client.stopped
    assert not client.connected
    assert not client.reconnecting
    assert not client.auto_reconnecting

    client.stop_listen()

    assert client.stopped
    assert not client.connected
    assert not client.reconnecting
    assert not client.auto_reconnecting


@patch_socket
async def test_auto_reconnect_loop(client: SocketClient):
    client.reconnect_interval = 0.01

    first_call = True

    async def _reconnect_nowait(*args, **kwargs):  # pylint: disable=unused-arguments
        if not first_call:
            assert client.auto_reconnecting

        client.connected = True
        client.reconnecting = False
        client.auto_reconnecting = False

    with patch(
        "pyaprilaire.socket_client.SocketClient._reconnect", new=_reconnect_nowait
    ):
        await client.start_listen()
        await client._auto_reconnect_loop()

    assert not client.stopped
    assert client.connected
    assert not client.reconnecting
    assert not client.auto_reconnecting


@patch_socket
async def test_auto_reconnect_loop_cancelled(client: SocketClient):
    client.reconnect_interval = 0.01

    async def _reconnect_nowait(self: SocketClient):
        self.connected = True
        self.reconnecting = False

    wait_for_mock = AsyncMock(side_effect=asyncio.exceptions.CancelledError)

    with patch(
        "pyaprilaire.socket_client.SocketClient._reconnect", new=_reconnect_nowait
    ), patch("asyncio.wait_for", new=wait_for_mock):
        await client.start_listen()
        await client._auto_reconnect_loop()

    assert not client.stopped
    assert client.connected
    assert not client.reconnecting


@patch_socket
async def test_auto_reconnect_loop_stopped(client: SocketClient):
    client.reconnect_interval = 0.01
    client.connected = False
    client.stopped = True

    await client._auto_reconnect_loop()

    assert client.stopped
    assert not client.connected
    assert not client.reconnecting


@patch_socket
async def test_cancel_auto_reconnect_loop(client: SocketClient):
    client.reconnect_interval = 1
    client.connected = True
    client.stopped = False

    async def cancel_auto_reconnect_loop():
        while not client.reconnect_break_future:
            await asyncio.sleep(0.01)
        client._cancel_auto_reconnect_loop()

    await asyncio.gather(cancel_auto_reconnect_loop(), client._auto_reconnect_loop())


@patch_socket
async def test_cancel_auto_reconnect_loop_state_error(client: SocketClient):
    loop = asyncio.get_event_loop()
    future = loop.create_future()

    client.reconnect_break_future = future

    future.set_result(None)

    client._cancel_auto_reconnect_loop()


@patch_socket
async def test_reconnect(client: SocketClient):
    client.stopped = False

    sleep_mock = AsyncMock()

    with patch("asyncio.sleep", new=sleep_mock):
        await client._reconnect(10)

    assert sleep_mock.await_count == 1
    assert not client.stopped
    assert client.connected
    assert not client.reconnecting


async def test_reconnect_exception(client: SocketClient):
    client.stopped = False

    sleep_mock = AsyncMock()

    def create_connection(*args, **kwargs):  # pylint: disable=unused-arguments
        client.cancelled = True
        raise Exception("Test failure")  # pylint: disable=broad-exception-raised

    create_connection_mock = AsyncMock(side_effect=create_connection)

    with (
        patch("asyncio.sleep", new=sleep_mock),
        patch("asyncio.BaseEventLoop.create_connection", new=create_connection_mock),
    ):
        await client._reconnect(10)

    assert sleep_mock.await_count == 2
    assert not client.stopped
    assert not client.connected
    assert client.reconnecting
    assert not client.auto_reconnecting


@patch_socket
async def test_reconnect_reconnecting(client: SocketClient):
    client.stopped = False
    client.connected = True
    client.reconnecting = True

    sleep_mock = AsyncMock()

    with patch("asyncio.sleep", new=sleep_mock):
        await client._reconnect(10)

    assert sleep_mock.await_count == 0
    assert not client.stopped
    assert client.connected
    assert client.reconnecting
